import copy
from typing import Callable, Generic, List, Optional, Tuple, TypeVar

T = TypeVar("T")


class Iterator(Generic[T]):
    @property
    def count(self) -> int:
        if self._total_count is not None:
            return self._total_count
        else:
            self._retrieve_next_page()
            return self._total_count  # type: ignore

    def __init__(
        self,
        *,
        results_fetcher: Callable[[int], Tuple[List[T], bool, int]],
        limit: int,
        current_page: int = 0,
        total_results_fetched: int = 0,
        results: Optional[List[T]] = None,
        total_count: Optional[int] = None,
        has_next: bool = True
    ):
        self._results_fetcher = results_fetcher
        self._limit = limit

        self._total_results_iterated = 0
        self._current_results_index = 0

        self._current_page = current_page
        self._total_results_fetched = total_results_fetched
        self._results: List[T] = [] if results is None else results
        self._total_count: Optional[int] = total_count
        self._has_next: bool = has_next

    def _retrieve_next_page(self) -> None:
        self._current_page += 1
        results, has_next, count = self._results_fetcher(self._current_page)

        self._total_results_fetched += len(results)
        self._current_results_index = 0
        self._results = results
        self._total_count = count
        self._has_next = has_next

    def __len__(self) -> int:
        return self.count if self.count < self._limit else self._limit

    def __iter__(self) -> "Iterator[T]":
        return type(self)(
            results_fetcher=self._results_fetcher,
            limit=self._limit,
            current_page=self._current_page,
            total_results_fetched=self._total_results_fetched,
            results=copy.deepcopy(self._results),
            total_count=self._total_count,
            has_next=self._has_next,
        )

    def __next__(self) -> T:
        if self._total_results_iterated >= self._limit:
            raise StopIteration()

        self._total_results_iterated += 1
        if self._total_results_iterated > self._total_results_fetched and self._has_next:
            self._retrieve_next_page()
        elif self._total_results_iterated > self._total_results_fetched:
            raise StopIteration()

        self._current_results_index += 1
        return self._results[self._current_results_index - 1]



import unittest
from datetime import datetime, timedelta, timezone

from caplena.helpers import Helpers


class HelperTests(unittest.TestCase):
    def test_user_agent_succeeds(self) -> None:
        user_agent = Helpers.get_user_agent(identifier="requests")
        self.assertRegex(user_agent, r"requests/0\.0\.\d python/3\.\d\.\d")

    def test_from_rfc3339_datetime_succeeds(self) -> None:
        dates = [
            "2021-12-29T14:46:55.787Z",
            "2021-09-01T11:43:49.508Z",
            "2021-09-01T11:43:49.500+04:00",
            "2021-12-31T08:43:49.444-07:00",
        ]
        expected = [
            datetime(2021, 12, 29, 14, 46, 55, 787000, tzinfo=timezone.utc).timestamp(),
            datetime(2021, 9, 1, 11, 43, 49, 508000, tzinfo=timezone.utc).timestamp(),
            datetime(
                2021, 9, 1, 11, 43, 49, 500000, tzinfo=timezone(timedelta(hours=4))
            ).timestamp(),
            datetime(
                2021, 12, 31, 8, 43, 49, 444000, tzinfo=timezone(timedelta(hours=-7))
            ).timestamp(),
        ]

        self.assertListEqual(
            expected, [Helpers.from_rfc3339_datetime(date).timestamp() for date in dates]
        )

    def test_to_rfc3339_datetime_succeeds(self) -> None:
        dates = [
            datetime(2021, 12, 29, 14, 46, 55, 787000, tzinfo=timezone.utc),
            datetime(2021, 9, 1, 11, 43, 49, 508000, tzinfo=timezone.utc),
            datetime(2021, 9, 1, 11, 43, 49, 500000, tzinfo=timezone(timedelta(hours=4))),
            datetime(2021, 12, 31, 8, 43, 49, 444000, tzinfo=timezone(timedelta(hours=-7))),
        ]
        expected = [
            "2021-12-29T14:46:55.787Z",
            "2021-09-01T11:43:49.508Z",
            "2021-09-01T11:43:49.500+04:00",
            "2021-12-31T08:43:49.444-07:00",
        ]

        self.assertListEqual(expected, [Helpers.to_rfc3339_datetime(date) for date in dates])

    def test_escaping_filter_string_succeeds(self) -> None:
        unescaped = [
            "abc cdef \\ \n ghj xyz",
            "th:is;is;just,a,ve:ry;lo:ng;tag",
            "just\\some\\\nmany\\\\\\backslashes\\n",
        ]
        escaped = [
            "abc cdef \\\\ \n ghj xyz",
            "th\\:is\\;is\\;just\\,a\\,ve\\:ry\\;lo\\:ng\\;tag",
            "just\\\\some\\\\\nmany\\\\\\\\\\\\backslashes\\\\n",
        ]

        self.assertListEqual(escaped, [Helpers.build_escaped_filter_str(un) for un in unescaped])



import json
from typing import Any, ClassVar, Optional


class ApiException(Exception):
    DEFAULT_MESSAGE: ClassVar[
        str
    ] = "An unknown error occurred. Please reach out to us at support@caplena.com."

    def __init__(
        self,
        *,
        type: str,
        code: str,
        message: str = DEFAULT_MESSAGE,
        details: Optional[str] = None,
        help: Optional[str] = None,
        context: Any = None,
    ):
        msg = f"{type}[{code}]: {message}"
        if details:
            msg += " " + details
        if help:
            msg += f" For more information, please visit {help}."
        if context:
            msg += f" (context={json.dumps(context)})"
        super().__init__(msg)

        self.type = type
        self.code = code
        self.message = message
        self.details = details
        self.help = help
        self.context = context



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from typing_extensions import Literal

from caplena.api import ApiOrdering
from caplena.endpoints.base_endpoint import BaseController, BaseObject, BaseResource
from caplena.filters.projects_filter import ProjectsFilter, RowsFilter
from caplena.helpers import Helpers
from caplena.http.http_response import HttpResponse
from caplena.iterator import Iterator

# --- Controller --- #


class ProjectsController(BaseController):
    def create(
        self,
        *,
        name: str,
        language: str,
        columns: List[Dict[str, Any]],
        tags: Optional[List[str]] = None,
        translation_engine: Optional[str] = None,
    ) -> "ProjectDetail":
        json = self.api.build_payload(
            name=name,
            language=language,
            columns=columns,
            tags=tags,
            translation_engine=translation_engine,
        )

        response = self.post(path="/projects", json=json)
        return self.build_response(response, resource=ProjectDetail)

    def retrieve(self, *, id: str) -> "ProjectDetail":
        response = self.get(path="/projects/{id}", path_params={"id": id})
        return self.build_response(response, resource=ProjectDetail)

    def remove(self, *, id: str) -> None:
        self.delete(path="/projects/{id}", path_params={"id": id})

    def list(
        self,
        *,
        limit: int = 10,
        filter: Optional[ProjectsFilter] = None,
        order_by: ApiOrdering = ApiOrdering.desc("last_modified"),
    ) -> "Iterator[ProjectList]":
        def fetcher(page: int) -> HttpResponse:
            return self.get(
                path="/projects",
                query_params={
                    "page": str(page),
                    "limit": str(limit),
                },
                filter=filter,
                order_by=order_by,
            )

        return self.build_iterator(fetcher=fetcher, limit=limit, resource=ProjectList)

    def append_rows(
        self,
        *,
        id: str,
        rows: List[Dict[str, Any]],
    ) -> "RowsAppend":
        response = self.post(
            path="/projects/{id}/rows/bulk",
            path_params={"id": id},
            json=rows,
            allowed_codes={202},
        )

        return self.build_response(response, resource=RowsAppend)

    def append_row(
        self,
        *,
        id: str,
        columns: List[Dict[str, Any]],
    ) -> "Row":
        json = self.api.build_payload(columns=columns)
        response = self.post(
            path="/projects/{id}/rows",
            path_params={"id": id},
            json=json,
        )

        return self.build_response(response, resource=Row)

    def list_rows(
        self,
        *,
        id: str,
        limit: int = 10,
        filter: Optional[RowsFilter] = None,
    ) -> "Iterator[Row]":
        def fetcher(page: int) -> HttpResponse:
            return self.get(
                path="/projects/{id}/rows",
                path_params={"id": id},
                query_params={
                    "page": str(page),
                    "limit": str(limit),
                },
                filter=filter,
            )

        return self.build_iterator(fetcher=fetcher, limit=limit, resource=Row)

    def retrieve_row(self, *, p_id: str, r_id: str) -> "Row":
        response = self.get(
            path="/projects/{p_id}/rows/{r_id}",
            path_params={"p_id": p_id, "r_id": r_id},
        )
        return self.build_response(response, resource=Row)


# --- Resources & Objects--- #


class ProjectDetail(BaseResource[ProjectsController]):
    class Column(BaseObject[ProjectsController]):
        __fields__ = {"ref", "name", "type"}

        ref: str
        name: str
        type: Literal["numerical", "boolean", "text", "date", "any", "text_to_analyze"]

    class TextToAnalyze(Column):
        class Topic(BaseResource[ProjectsController]):
            class Sentiment(BaseObject[ProjectsController]):
                __fields__ = {"code", "label"}

                code: int
                label: str

            __fields__ = {
                "label",
                "category",
                "color",
                "description",
                "sentiment_enabled",
                "sentiment_neutral",
                "sentiment_negative",
                "sentiment_positive",
            }

            label: str
            category: str
            color: str
            description: str
            sentiment_enabled: bool
            sentiment_neutral: Sentiment
            sentiment_negative: Sentiment
            sentiment_positive: Sentiment

            @classmethod
            def parse_obj(cls, obj: Dict[str, Any]) -> "ProjectDetail.TextToAnalyze.Topic":
                obj["sentiment_neutral"] = cls.Sentiment.parse_obj(obj["sentiment_neutral"])
                obj["sentiment_negative"] = cls.Sentiment.parse_obj(obj["sentiment_negative"])
                obj["sentiment_positive"] = cls.Sentiment.parse_obj(obj["sentiment_positive"])

                return super().parse_obj(obj)

        class Metadata(BaseObject[ProjectsController]):
            __fields__ = {"reviewed_count"}

            reviewed_count: int

        __fields__ = {"ref", "name", "type", "description", "topics", "metadata"}

        type: Literal["text_to_analyze"]
        description: str
        topics: List[Topic]
        metadata: Metadata

        @classmethod
        def parse_obj(cls, obj: Dict[str, Any]) -> "ProjectDetail.TextToAnalyze":
            obj["topics"] = [cls.Topic.parse_obj(topic) for topic in obj["topics"]]
            obj["metadata"] = cls.Metadata.parse_obj(obj["metadata"])
            return super().parse_obj(obj)

    class Auxiliary(Column):
        type: Literal["numerical", "boolean", "text", "date", "any"]

    __fields__ = {
        "name",
        "owner",
        "tags",
        "upload_status",
        "language",
        "columns",
        "created",
        "last_modified",
        "translation_status",
        "translation_engine",
    }

    name: str
    owner: str
    tags: List[str]
    upload_status: Literal["pending", "in_progress", "succeeded", "failed"]
    # fmt: off
    language: Literal["af", "sq", "eu", "ca", "cs", "da", "nl", "en", "et", "fi",
                      "fr", "gl", "de", "el", "hu", "is", "it", "lb", "lt", "lv", "mk", "no",
                      "pl", "pt", "ro", "sr", "sk", "sl", "es", "sv", "tr"]
    # fmt: on
    columns: List[Column]
    created: datetime
    last_modified: datetime
    translation_status: Optional[str]
    translation_engine: Optional[str]

    def remove(self) -> None:
        self.controller.remove(id=self.id)

    def append_row(self, *, columns: List[Dict[str, Any]]) -> "Row":
        return self.controller.append_row(id=self.id, columns=columns)

    def append_rows(self, *, rows: List[Dict[str, Any]]) -> "RowsAppend":
        return self.controller.append_rows(id=self.id, rows=rows)

    def list_rows(
        self,
        *,
        limit: int = 10,
        filter: Optional[RowsFilter] = None,
    ) -> "Iterator[Row]":
        return self.controller.list_rows(id=self.id, limit=limit, filter=filter)

    def retrieve_row(self, *, id: str) -> "Row":
        return self.controller.retrieve_row(p_id=self.id, r_id=id)

    def refresh(self) -> None:
        project = self.controller.retrieve(id=self.id)
        self.refresh_from(attrs=project._attrs)

    @classmethod
    def parse_obj(cls, obj: Dict[str, Any]) -> "ProjectDetail":
        obj["columns"] = [
            cls.TextToAnalyze.parse_obj(column)
            if column["type"] == "text_to_analyze"
            else cls.Auxiliary.parse_obj(column)
            for column in obj["columns"]
        ]
        obj["created"] = Helpers.from_rfc3339_datetime(obj["created"])
        obj["last_modified"] = Helpers.from_rfc3339_datetime(obj["last_modified"])

        return super().parse_obj(obj)


class ProjectList(BaseResource[ProjectsController]):
    __fields__ = {
        "name",
        "owner",
        "tags",
        "upload_status",
        "language",
        "created",
        "last_modified",
        "translation_status",
        "translation_engine",
    }

    name: str
    owner: str
    tags: List[str]
    upload_status: Literal["pending", "in_progress", "succeeded", "failed"]
    # fmt: off
    language: Literal["af", "sq", "eu", "ca", "cs", "da", "nl", "en", "et", "fi",
                      "fr", "gl", "de", "el", "hu", "is", "it", "lb", "lt", "lv", "mk", "no",
                      "pl", "pt", "ro", "sr", "sk", "sl", "es", "sv", "tr"]
    # fmt: on
    created: datetime
    last_modified: datetime
    translation_status: Optional[str]
    translation_engine: Optional[str]

    def remove(self) -> None:
        self.controller.remove(id=self.id)

    def append_row(self, *, columns: List[Dict[str, Any]]) -> "Row":
        return self.controller.append_row(id=self.id, columns=columns)

    def append_rows(self, *, rows: List[Dict[str, Any]]) -> "RowsAppend":
        return self.controller.append_rows(id=self.id, rows=rows)

    def list_rows(
        self,
        *,
        limit: int = 10,
        filter: Optional[RowsFilter] = None,
    ) -> "Iterator[Row]":
        return self.controller.list_rows(id=self.id, limit=limit, filter=filter)

    def retrieve_row(self, *, id: str) -> "Row":
        return self.controller.retrieve_row(p_id=self.id, r_id=id)

    def refresh(self) -> None:
        project = self.controller.retrieve(id=self.id)
        self.refresh_from(attrs=project._attrs)

    @classmethod
    def parse_obj(cls, obj: Dict[str, Any]) -> "ProjectList":
        obj["created"] = Helpers.from_rfc3339_datetime(obj["created"])
        obj["last_modified"] = Helpers.from_rfc3339_datetime(obj["last_modified"])

        return super().parse_obj(obj)


class RowsAppend(BaseObject[ProjectsController]):
    __fields__ = {"status", "queued_rows_count", "estimated_minutes"}

    status: Literal["pending"]
    queued_rows_count: int
    estimated_minutes: float


class Row(BaseResource[ProjectsController]):
    class Column(BaseObject[ProjectsController]):
        __fields__ = {"ref", "type", "value"}

        ref: str
        type: Literal["numerical", "boolean", "text", "date", "any", "text_to_analyze"]
        value: Union[float, bool, None, str, datetime]

    class NumericalColumn(Column):
        type: Literal["numerical"]
        value: Optional[float]

    class BooleanColumn(Column):
        type: Literal["boolean"]
        value: Optional[bool]

    class DateColumn(Column):
        type: Literal["date"]
        value: Optional[datetime]

        @classmethod
        def parse_obj(cls, obj: Dict[str, Any]) -> "Row.DateColumn":
            if obj["value"] is not None:
                obj["value"] = Helpers.from_rfc3339_datetime(obj["value"])

            return super().parse_obj(obj)

    class AnyColumn(Column):
        type: Literal["any"]
        value: None

    class TextColumn(Column):
        type: Literal["text"]
        value: str

    class TextToAnalyzeColumn(Column):
        class Topic(BaseObject[ProjectsController]):
            __fields__ = {"id", "label", "category", "code", "sentiment_label", "sentiment"}

            id: str
            label: str
            category: str
            code: int
            sentiment_label: str
            sentiment: Literal["neutral", "positive", "negative"]

        __fields__ = {
            "ref",
            "type",
            "value",
            "was_reviewed",
            "sentiment_overall",
            "source_language",
            "translated_value",
            "topics",
        }

        type: Literal["text_to_analyze"]
        value: str
        was_reviewed: Optional[bool]
        sentiment_overall: Optional[Literal["neutral", "positive", "negative"]]
        # fmt: off
        source_language: Optional[Literal["af", "am", "ar", "az", "be", "bg", "bn", "bs", "ca", "ceb", "co", "cs", "cy", "da",
                                          "de", "el", "en", "eo", "es", "et", "eu", "fa", "fi", "fr", "fy", "ga", "gd", "gl",
                                          "gu", "ha", "haw", "hi", "hmn", "hr", "ht", "hu", "hy", "id", "ig", "is", "it", "iw",
                                          "he", "ja", "jw", "ka", "kk", "km", "kn", "ko", "ku", "ky", "la", "lb", "lo", "lt",
                                          "lv", "mg", "mi", "mk", "ml", "mn", "mr", "ms", "mt", "my", "ne", "nl", "no", "ny",
                                          "pa", "pl", "ps", "pt", "ro", "ru", "sd", "si", "sk", "sl", "sm", "sn", "so", "sq",
                                          "sr", "st", "su", "sv", "sw", "ta", "te", "tg", "th", "tl", "tr", "tk", "uk", "ur",
                                          "uz", "vi", "xh", "yi", "yo", "zh", "zh-CN", "zh-TW", "zu"]]
        # fmt: on
        translated_value: Optional[str]
        topics: List[Topic]

        @classmethod
        def parse_obj(cls, obj: Dict[str, Any]) -> "Row.TextToAnalyzeColumn":
            obj["topics"] = [cls.Topic.parse_obj(topic) for topic in obj["topics"]]
            return super().parse_obj(obj)

    __fields__ = {"created", "last_modified", "columns"}

    created: datetime
    last_modified: datetime
    columns: List[Column]

    @classmethod
    def parse_obj(cls, obj: Dict[str, Any]) -> "Row":
        type_to_column = {
            "numerical": cls.NumericalColumn,
            "boolean": cls.BooleanColumn,
            "date": cls.DateColumn,
            "any": cls.AnyColumn,
            "text": cls.TextColumn,
            "text_to_analyze": cls.TextToAnalyzeColumn,
        }
        obj["columns"] = [
            type_to_column[column["type"]].parse_obj(column) for column in obj["columns"]
        ]
        obj["created"] = Helpers.from_rfc3339_datetime(obj["created"])
        obj["last_modified"] = Helpers.from_rfc3339_datetime(obj["last_modified"])

        return super().parse_obj(obj)



from datetime import datetime

from caplena.api import ApiFilter, ZeroOrMany


class ProjectsFilter(ApiFilter):
    @classmethod
    def name(
        cls,
        *,
        exact__i: ZeroOrMany[str] = None,
        contains__i: ZeroOrMany[str] = None,
    ) -> "ProjectsFilter":
        return cls.construct(
            name="name",
            filters={
                "exact.i": exact__i,
                "contains.i": contains__i,
            },
        )

    @classmethod
    def owner(
        cls,
        *,
        id: ZeroOrMany[str],
        email__exact__i: ZeroOrMany[str],
        email__contains__i: ZeroOrMany[str],
    ) -> "ProjectsFilter":
        return cls.construct(
            name="owner",
            filters={
                "id": id,
                "email.exact.i": email__exact__i,
                "email.contains.i": email__contains__i,
            },
        )

    @classmethod
    def tags(cls, tag: ZeroOrMany[str]) -> "ProjectsFilter":
        return cls.construct(
            name="tags",
            filters={
                cls.DEFAULT: tag,
            },
        )

    @classmethod
    def upload_status(cls, upload_status: ZeroOrMany[str]) -> "ProjectsFilter":
        return cls.construct(
            name="upload_status",
            filters={
                cls.DEFAULT: upload_status,
            },
        )

    @classmethod
    def language(cls, language: ZeroOrMany[str]) -> "ProjectsFilter":
        return cls.construct(
            name="language",
            filters={
                cls.DEFAULT: language,
            },
        )

    @classmethod
    def translation_status(cls, translation_status: ZeroOrMany[str]) -> "ProjectsFilter":
        return cls.construct(
            name="translation_status",
            filters={
                cls.DEFAULT: translation_status,
            },
        )

    @classmethod
    def translation_engine(cls, translation_engine: ZeroOrMany[str]) -> "ProjectsFilter":
        return cls.construct(
            name="translation_engine",
            filters={
                cls.DEFAULT: translation_engine,
            },
        )

    @classmethod
    def created(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        gt: ZeroOrMany[datetime] = None,
        lte: ZeroOrMany[datetime] = None,
        lt: ZeroOrMany[datetime] = None,
        year: ZeroOrMany[int] = None,
        year__gte: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lte: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        month__gte: ZeroOrMany[int] = None,
        month__gt: ZeroOrMany[int] = None,
        month__lte: ZeroOrMany[int] = None,
        month__lt: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
        day__gte: ZeroOrMany[int] = None,
        day__gt: ZeroOrMany[int] = None,
        day__lte: ZeroOrMany[int] = None,
        day__lt: ZeroOrMany[int] = None,
    ) -> "ProjectsFilter":
        return cls.construct(
            name="created",
            filters={
                "gte": gte,
                "gt": gt,
                "lte": lte,
                "lt": lt,
                "year": year,
                "year.gte": year__gte,
                "year.gt": year__gt,
                "year.lte": year__lte,
                "year.lt": year__lt,
                "month": month,
                "month.gte": month__gte,
                "month.gt": month__gt,
                "month.lte": month__lte,
                "month.lt": month__lt,
                "day": day,
                "day.gte": day__gte,
                "day.gt": day__gt,
                "day.lte": day__lte,
                "day.lt": day__lt,
            },
        )

    @classmethod
    def last_modified(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        gt: ZeroOrMany[datetime] = None,
        lte: ZeroOrMany[datetime] = None,
        lt: ZeroOrMany[datetime] = None,
        year: ZeroOrMany[int] = None,
        year__gte: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lte: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        month__gte: ZeroOrMany[int] = None,
        month__gt: ZeroOrMany[int] = None,
        month__lte: ZeroOrMany[int] = None,
        month__lt: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
        day__gte: ZeroOrMany[int] = None,
        day__gt: ZeroOrMany[int] = None,
        day__lte: ZeroOrMany[int] = None,
        day__lt: ZeroOrMany[int] = None,
    ) -> "ProjectsFilter":
        return cls.construct(
            name="last_modified",
            filters={
                "gte": gte,
                "gt": gt,
                "lte": lte,
                "lt": lt,
                "year": year,
                "year.gte": year__gte,
                "year.gt": year__gt,
                "year.lte": year__lte,
                "year.lt": year__lt,
                "month": month,
                "month.gte": month__gte,
                "month.gt": month__gt,
                "month.lte": month__lte,
                "month.lt": month__lt,
                "day": day,
                "day.gte": day__gte,
                "day.gt": day__gt,
                "day.lte": day__lte,
                "day.lt": day__lt,
            },
        )


class RowsFilter(ApiFilter):
    class Columns:
        @staticmethod
        def numerical(
            *,
            ref: str,
            exact: ZeroOrMany[float] = None,
            gte: ZeroOrMany[float] = None,
            gt: ZeroOrMany[float] = None,
            lte: ZeroOrMany[float] = None,
            lt: ZeroOrMany[float] = None,
        ) -> "RowsFilter":
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[numerical]": exact,
                    f"{ref}[numerical].gte": gte,
                    f"{ref}[numerical].gt": gt,
                    f"{ref}[numerical].lte": lte,
                    f"{ref}[numerical].lt": lt,
                },
            )

        @classmethod
        def boolean(
            cls,
            *,
            ref: str,
            exact: ZeroOrMany[bool] = None,
        ) -> "RowsFilter":
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[boolean]": exact,
                },
            )

        @classmethod
        def text(
            cls,
            *,
            ref: str,
            exact__i: ZeroOrMany[str] = None,
            contains__i: ZeroOrMany[str] = None,
        ) -> "RowsFilter":
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[text].exact.i": exact__i,
                    f"{ref}[text].contains.i": contains__i,
                },
            )

        @classmethod
        def text_to_analyze(
            cls,
            *,
            ref: str,
            exact__i: ZeroOrMany[str] = None,
            contains__i: ZeroOrMany[str] = None,
            was_reviewed: ZeroOrMany[bool] = None,
            source_language: ZeroOrMany[str] = None,
            translated_value__exact__i: ZeroOrMany[str] = None,
            translated_value__contains__i: ZeroOrMany[str] = None,
        ) -> "RowsFilter":
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[text_to_analyze].exact.i": exact__i,
                    f"{ref}[text_to_analyze].contains.i": contains__i,
                    f"{ref}[text_to_analyze].was_reviewed": was_reviewed,
                    f"{ref}[text_to_analyze].source_language": source_language,
                    f"{ref}[text_to_analyze].translated_value.exact.i": translated_value__exact__i,
                    f"{ref}[text_to_analyze].translated_value.contains.i": translated_value__contains__i,
                },
            )

    @classmethod
    def created(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        gt: ZeroOrMany[datetime] = None,
        lte: ZeroOrMany[datetime] = None,
        lt: ZeroOrMany[datetime] = None,
        year: ZeroOrMany[int] = None,
        year__gte: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lte: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        month__gte: ZeroOrMany[int] = None,
        month__gt: ZeroOrMany[int] = None,
        month__lte: ZeroOrMany[int] = None,
        month__lt: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
        day__gte: ZeroOrMany[int] = None,
        day__gt: ZeroOrMany[int] = None,
        day__lte: ZeroOrMany[int] = None,
        day__lt: ZeroOrMany[int] = None,
    ) -> "RowsFilter":
        return cls.construct(
            name="created",
            filters={
                "gte": gte,
                "gt": gt,
                "lte": lte,
                "lt": lt,
                "year": year,
                "year.gte": year__gte,
                "year.gt": year__gt,
                "year.lte": year__lte,
                "year.lt": year__lt,
                "month": month,
                "month.gte": month__gte,
                "month.gt": month__gt,
                "month.lte": month__lte,
                "month.lt": month__lt,
                "day": day,
                "day.gte": day__gte,
                "day.gt": day__gt,
                "day.lte": day__lte,
                "day.lt": day__lt,
            },
        )

    @classmethod
    def last_modified(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        gt: ZeroOrMany[datetime] = None,
        lte: ZeroOrMany[datetime] = None,
        lt: ZeroOrMany[datetime] = None,
        year: ZeroOrMany[int] = None,
        year__gte: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lte: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        month__gte: ZeroOrMany[int] = None,
        month__gt: ZeroOrMany[int] = None,
        month__lte: ZeroOrMany[int] = None,
        month__lt: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
        day__gte: ZeroOrMany[int] = None,
        day__gt: ZeroOrMany[int] = None,
        day__lte: ZeroOrMany[int] = None,
        day__lt: ZeroOrMany[int] = None,
    ) -> "RowsFilter":
        return cls.construct(
            name="last_modified",
            filters={
                "gte": gte,
                "gt": gt,
                "lte": lte,
                "lt": lt,
                "year": year,
                "year.gte": year__gte,
                "year.gt": year__gt,
                "year.lte": year__lte,
                "year.lt": year__lt,
                "month": month,
                "month.gte": month__gte,
                "month.gt": month__gt,
                "month.lte": month__lte,
                "month.lt": month__lt,
                "day": day,
                "day.gte": day__gte,
                "day.gt": day__gt,
                "day.lte": day__lte,
                "day.lt": day__lt,
            },
        )



from typing import Iterable, Type, Union

from caplena.api import ApiBaseUri, ApiVersion
from caplena.configuration import Configuration
from caplena.endpoints.projects_endpoint import ProjectsController
from caplena.http.http_client import HttpClient, HttpMethod, HttpRetry
from caplena.http.requests_http_client import RequestsHttpClient
from caplena.logging.logger import LoggingLevel


class Client:
    @property
    def projects(self) -> ProjectsController:
        return self._projects_controller

    @property
    def config(self) -> Configuration:
        return self._config

    def __init__(
        self,
        api_key: str,
        *,
        api_base_uri: ApiBaseUri = ApiBaseUri.PRODUCTION,
        api_version: ApiVersion = ApiVersion.DEFAULT,
        timeout: int = HttpClient.DEFAULT_TIMEOUT,
        max_retries: int = HttpRetry.DEFAULT_MAX_RETRIES,
        backoff_factor: float = HttpRetry.DEFAULT_BACKOFF_FACTOR,
        retry_status_codes: Iterable[int] = HttpRetry.DEFAULT_STATUS_CODES_TO_RETRY,
        retry_methods: Iterable[HttpMethod] = HttpRetry.DEFAULT_ALLOWED_METHOD,
        http_client: Union[Type[HttpClient], HttpClient] = RequestsHttpClient,
        logging_level: LoggingLevel = LoggingLevel.WARNING,
    ):
        self._config = Configuration(
            api_key=api_key,
            http_client=http_client,
            api_base_uri=api_base_uri,
            api_version=api_version,
            timeout=timeout,
            max_retries=max_retries,
            backoff_factor=backoff_factor,
            retry_status_codes=retry_status_codes,
            retry_methods=retry_methods,
            logging_level=logging_level,
        )

        self._projects_controller = ProjectsController(config=self._config)



import unittest
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, cast

from caplena.api.api_exception import ApiException
from caplena.endpoints.projects_endpoint import ProjectDetail, ProjectsController, Row
from caplena.filters.projects_filter import ProjectsFilter, RowsFilter
from tests.common import common_config


def project_create_payload() -> Dict[str, Any]:
    return {
        "name": "Project Name",
        "tags": ["my-tag"],
        "language": "en",
        "translation_engine": None,
        "columns": [
            {
                "ref": "customer_age",
                "type": "numerical",
                "name": "Age of the customer",
            },
            {
                "ref": "our_strengths",
                "name": "Do you like us?",
                "type": "text_to_analyze",
                "description": "Please explain what our strengths are.",
                "topics": [
                    {
                        "label": "Some Code Label",
                        "sentiment_enabled": False,
                        "category": "SOME_CATEGORY",
                    },
                    {
                        "label": "Another Code Label",
                        "sentiment_enabled": False,
                        "category": "ANOTHER_CATEGORY",
                    },
                ],
            },
        ],
    }


def project_rows_create_payload() -> List[Dict[str, Any]]:
    return [
        {
            "columns": [
                {"ref": "customer_age", "value": 120},
                {"ref": "our_strengths", "value": "This is nice."},
            ]
        },
        {
            "columns": [
                {"ref": "customer_age", "value": None},
                {"ref": "our_strengths", "value": "Some other text."},
            ]
        },
    ]


class ProjectsControllerTests(unittest.TestCase):
    controller: ClassVar[ProjectsController]

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.controller = ProjectsController(config=common_config)

    def setUp(self) -> None:
        self.created_projects: List[str] = []

    def tearDown(self) -> None:
        for project_id in self.created_projects:
            try:
                self.controller.remove(id=project_id)
            except ApiException:
                print("Could not remove project with id:", project_id)

    def create_project(self, *, payload: Optional[Dict[str, Any]] = None) -> ProjectDetail:
        if payload is None:
            payload = project_create_payload()

        project = self.controller.create(**payload)
        self.created_projects.append(project.id)

        return project

    def test_creating_a_project_succeeds(self) -> None:
        project = self.create_project()

        self.assertIsInstance(project.id, str)
        self.assertEqual("Project Name", project.name)
        self.assertIsInstance(project.owner, str)
        self.assertListEqual(["my-tag"], project.tags)
        self.assertEqual("pending", project.upload_status)
        self.assertEqual("en", project.language)
        self.assertEqual(None, project.translation_status)
        self.assertEqual(None, project.translation_engine)
        self.assertIsInstance(project.created, datetime)
        self.assertIsInstance(project.last_modified, datetime)

        self.assertEqual(2, len(project.columns))
        self.assertIsInstance(project.columns[0], ProjectDetail.TextToAnalyze)
        self.assertIsInstance(project.columns[1], ProjectDetail.Auxiliary)
        our_strengths = cast(ProjectDetail.TextToAnalyze, project.columns[0])
        customer_age = cast(ProjectDetail.Auxiliary, project.columns[1])

        self.assertEqual("our_strengths", our_strengths.ref)
        self.assertEqual("Do you like us?", our_strengths.name)
        self.assertEqual("text_to_analyze", our_strengths.type)
        self.assertEqual("Please explain what our strengths are.", our_strengths.description)
        self.assertDictEqual({"reviewed_count": 0}, our_strengths.metadata.dict())
        self.assertEqual(2, len(our_strengths.topics))

        topic1, topic2 = our_strengths.topics[0], our_strengths.topics[1]
        self.assertRegex(topic1.id, r"^cd_")
        self.assertEqual("Some Code Label", topic1.label)
        self.assertEqual("SOME_CATEGORY", topic1.category)
        self.assertEqual("", topic1.color)
        self.assertEqual("", topic1.description)
        self.assertEqual(False, topic1.sentiment_enabled)
        self.assertDictEqual({"code": 0, "label": ""}, topic1.sentiment_neutral.dict())
        self.assertDictEqual({"code": -1, "label": ""}, topic1.sentiment_negative.dict())
        self.assertDictEqual({"code": -1, "label": ""}, topic1.sentiment_positive.dict())

        self.assertRegex(topic2.id, r"^cd_")
        self.assertEqual("Another Code Label", topic2.label)
        self.assertEqual("ANOTHER_CATEGORY", topic2.category)
        self.assertEqual("", topic2.color)
        self.assertEqual("", topic2.description)
        self.assertEqual(False, topic2.sentiment_enabled)
        self.assertDictEqual({"code": 1, "label": ""}, topic2.sentiment_neutral.dict())
        self.assertDictEqual({"code": -1, "label": ""}, topic2.sentiment_negative.dict())
        self.assertDictEqual({"code": -1, "label": ""}, topic2.sentiment_positive.dict())

        self.assertEqual("customer_age", customer_age.ref)
        self.assertEqual("Age of the customer", customer_age.name)
        self.assertEqual("numerical", customer_age.type)

    def test_retrieving_a_project_succeeds(self) -> None:
        project = self.create_project()
        retrieved = self.controller.retrieve(id=project.id)

        self.assertDictEqual(project.dict(), retrieved.dict())

    def test_removing_a_project_succeeds(self) -> None:
        old_num_projects = self.controller.list(limit=1).count
        project = self.create_project()
        interim_num_projects = self.controller.list(limit=1).count
        self.controller.remove(id=project.id)
        new_num_projects = self.controller.list(limit=1).count

        self.assertEqual(old_num_projects, new_num_projects)
        self.assertEqual(old_num_projects + 1, interim_num_projects)

    def test_listing_all_projects_succeeds(self) -> None:
        project = self.create_project()
        projects = self.controller.list(limit=1)

        self.assertIsInstance(projects.count, int)
        self.assertGreaterEqual(projects.count, 0)
        self.assertEqual(len(projects), 1)
        retrieved = (list(projects))[0]

        project_dict = project.dict()
        project_dict.pop("columns")
        self.assertDictEqual(project_dict, retrieved.dict())

    def test_filtering_projects_succeeds(self) -> None:
        filt = ProjectsFilter.language("tr")
        projects = self.controller.list(filter=filt)

        self.assertEqual(0, len(projects))
        self.assertEqual(0, projects.count)

    def test_appending_multiple_rows_succeeds(self) -> None:
        project = self.create_project()
        result = self.controller.append_rows(id=project.id, rows=project_rows_create_payload())

        self.assertEqual("pending", result.status)
        self.assertEqual(2, result.queued_rows_count)
        self.assertEqual(1.02, result.estimated_minutes)

    def test_appending_single_row_succeeds(self) -> None:
        project = self.create_project()
        columns: List[Dict[str, Any]] = [
            {"ref": "customer_age", "value": None},
            {"ref": "our_strengths", "value": "Some other text."},
        ]
        row = self.controller.append_row(id=project.id, columns=columns)

        self.assertIsInstance(row.id, str)
        self.assertIsInstance(row.created, datetime)
        self.assertIsInstance(row.last_modified, datetime)

        self.assertEqual(2, len(row.columns))
        self.assertIsInstance(row.columns[0], Row.TextToAnalyzeColumn)
        self.assertIsInstance(row.columns[1], Row.NumericalColumn)
        our_strengths = cast(Row.TextToAnalyzeColumn, row.columns[0])
        customer_age = cast(Row.NumericalColumn, row.columns[1])

        self.assertEqual("our_strengths", our_strengths.ref)
        self.assertEqual("text_to_analyze", our_strengths.type)
        self.assertEqual("Some other text.", our_strengths.value)
        self.assertEqual(False, our_strengths.was_reviewed)
        self.assertEqual(None, our_strengths.source_language)
        self.assertEqual(None, our_strengths.translated_value)
        self.assertEqual(1, len(our_strengths.topics))
        topic = our_strengths.topics[0]
        self.assertRegex(topic.id, r"^cd_")
        # TODO: what should be returned here?

        self.assertEqual("customer_age", customer_age.ref)
        self.assertEqual("numerical", customer_age.type)
        self.assertEqual(None, customer_age.value)

    def test_listing_all_rows_succeeds(self) -> None:
        project = self.create_project()
        row1 = self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": None},
                {"ref": "our_strengths", "value": "Some other text."},
            ],
        )
        row2 = self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": 12},
                {"ref": "our_strengths", "value": "This is my review. Very nice."},
            ],
        )
        rows = [row2.dict(), row1.dict()]
        retrieved = self.controller.list_rows(id=project.id, limit=2)

        self.assertEqual(2, retrieved.count)
        self.assertEqual(2, len(retrieved))

        retrieved_dict = [row.dict() for row in retrieved]
        self.assertListEqual(rows, retrieved_dict)

    def test_filtering_rows_succeeds(self) -> None:
        project = self.create_project()
        self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": None},
                {"ref": "our_strengths", "value": "Some other text."},
            ],
        )

        with_results = self.controller.list_rows(id=project.id)
        self.assertEqual(1, with_results.count)

        filt = RowsFilter.Columns.text_to_analyze(ref="our_strengths", contains__i=["some"])
        filtered_results = self.controller.list_rows(id=project.id, filter=filt)
        self.assertEqual(1, filtered_results.count)

        filt = RowsFilter.Columns.text_to_analyze(ref="our_strengths", was_reviewed=True)
        no_results = self.controller.list_rows(id=project.id, filter=filt)
        self.assertEqual(0, no_results.count)

    def test_retreiving_a_row_succeeds(self) -> None:
        project = self.create_project()
        row = self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": 400},
                {"ref": "our_strengths", "value": "Some other text."},
            ],
        )
        retrieved = self.controller.retrieve_row(p_id=project.id, r_id=row.id)

        self.assertDictEqual(row.dict(), retrieved.dict())



from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from caplena.api import ApiFilter, ApiOrdering
from caplena.api.api_requestor import ApiRequestor
from caplena.configuration import Configuration
from caplena.helpers import Helpers
from caplena.http.http_response import HttpResponse
from caplena.iterator import Iterator

OBO = TypeVar("OBO", bound="BaseObject[Any]")
BO = TypeVar("BO", bound="BaseObject[Any]")
BC = TypeVar("BC", bound="BaseController")
T = TypeVar("T")


class BaseController:
    DEFAULT_ALLOWED_CODES: ClassVar[Iterable[int]] = frozenset({200})
    DEFAULT_ALLOWED_POST_CODES: ClassVar[Iterable[int]] = frozenset({201})
    DEFAULT_ALLOWED_DELETE_CODES: ClassVar[Iterable[int]] = frozenset({204})

    @property
    def config(self) -> Configuration:
        return self._config

    @property
    def api(self) -> ApiRequestor:
        return self._config.api_requestor

    def __init__(self, *, config: Configuration):
        self._config = config

    def get(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        filter: Optional[ApiFilter] = None,
        order_by: Optional[ApiOrdering] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.get(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            path_params=path_params,
            query_params=query_params,
            filter=filter,
            order_by=order_by,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def post(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_POST_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.post(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            path_params=path_params,
            query_params=query_params,
            json=json,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def put(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.put(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            path_params=path_params,
            query_params=query_params,
            json=json,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def patch(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.patch(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            path_params=path_params,
            query_params=query_params,
            json=json,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def delete(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_DELETE_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.delete(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            path_params=path_params,
            query_params=query_params,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def build_response(self, response: HttpResponse, *, resource: Type[BO]) -> BO:
        json = self._retrieve_json_or_raise(response)
        return resource.build_obj(obj=json, controller=self)

    def build_iterator(
        self,
        *,
        fetcher: Callable[[int], HttpResponse],
        limit: int,
        resource: Type[BO],
    ) -> Iterator[BO]:
        def results_fetcher(page: int) -> Tuple[List[BO], bool, int]:
            response = fetcher(page)
            json = self._retrieve_json_or_raise(response)

            results = [resource.build_obj(res, controller=self) for res in json["results"]]
            return results, json["next_url"] is not None, json["count"]

        return Iterator(
            results_fetcher=results_fetcher,
            limit=limit,
        )

    def _retrieve_json_or_raise(self, response: HttpResponse) -> Dict[str, Any]:
        json = response.json
        if json is None:
            raise self.api.build_exc(response)
        else:
            return json


class BaseObject(Generic[BC]):
    __fields__: ClassVar[Set[str]] = set()
    __mutable__: ClassVar[Set[str]] = set()

    _attrs: Dict[str, Any]
    _unpersisted_attributes: Set[str]
    _controller: Optional[BC]

    @property
    def controller(self) -> BC:
        if self._controller is None:
            raise ValueError(
                "You cannot access the non-existing controller for this object. HINT: This object either "
                "does not have a controller attached, or you forgot to manually set the controller after initializing it. "
                "(object.controller = your_controller)"
            )
        return self._controller

    @controller.setter
    def controller(self, value: BC) -> None:
        self._controller = value

        for field in self.__fields__:
            self._recursive_controller_set(self._attrs[field], value=value)

    def __init__(self, **attrs: Any):
        self._controller = None

        self.refresh_from(attrs=attrs)

    def _recursive_controller_set(self, attr: Any, *, value: BC) -> None:
        if isinstance(attr, BaseObject):
            self._controller = value
        elif isinstance(attr, list):
            [self._recursive_controller_set(i, value=value) for i in attr]  # type: ignore

    def refresh_from(self, *, attrs: Dict[str, Any]) -> None:
        # pick only allowed attributes
        partial = Helpers.partial_dict(attrs, self.__fields__)
        validated = partial

        self._attrs = validated
        self._unpersisted_attributes = set()

    def dict(self) -> Dict[str, Any]:
        resource: Dict[str, Any] = {}
        for field in self.__fields__:
            attr = self._attrs[field]
            resource[field] = self._recursive_dict(attr)

        return resource

    def _recursive_dict(self, attr: Any) -> Any:
        if isinstance(attr, BaseObject):
            return attr.dict()
        elif isinstance(attr, list):
            return [
                self._recursive_dict(i) for i in attr
            ]  # pyright: reportUnknownVariableType=false
        else:
            return attr

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.__fields__ and name in self.__mutable__:
            self._unpersisted_attributes.add(name)
            self._attrs[name] = value
        elif name in self.__fields__:
            raise AttributeError(
                f"{name}. HINT: You cannot modify this attribute, as it is immutable."
            )
        else:
            super().__setattr__(name, value)

    def __getattr__(self, name: str) -> Any:
        if name in self.__fields__:
            return self._attrs[name]
        else:
            return super().__getattribute__(name)

    def __delattr__(self, name: str) -> None:
        raise ValueError(f"{name}. HINT: You cannot delete any attributes.")

    @classmethod
    def build_obj(cls: Type[BO], obj: Dict[str, Any], *, controller: Optional[BC]) -> BO:
        instance = cls.parse_obj(obj)
        instance.controller = controller
        return instance

    @classmethod
    def parse_obj(cls: Type[BO], obj: Dict[str, Any]) -> BO:
        return cls(**obj)


class BaseResource(BaseObject[BC]):
    @property
    def id(self) -> str:
        return self._id

    def __init__(self, id: str, **attrs: Any):
        super().__init__(**attrs)
        self._id = id

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(id={self._id})"

    def dict(self) -> Dict[str, Any]:
        resource = super().dict()
        resource["id"] = self._id
        return resource
